#include <app/App.h>
#include <common/nodes/Node.h>
#include <common/nodes/NodeConnPool.h>
#include <common/nodes/Node.h>
#include <nodes/NodeStoreEx.h>
#include <toolkit/NoAllocBufferStore.h>
#include <toolkit/SignalTk.h>
#include "MessagingTk.h"


#define MSGTK_KMALLOC_RECV_BUF_LEN  4*1024 /* kmalloc-style recv is only ok for small replies */


/**
 * Send a request message to a node and receive the response.
 *
 * @param rrArgs:
 * .node receiver of msg;
 * .requestMsg the message that should be sent to the receiver;
 * .respMsgType expected response message type;
 * .outRespBuf response buffer if successful (must be returned to store by the caller);
 * .outRespMsg response message if successful (must be deleted by the caller);
 * @return _COMMUNICATION on comm error, _INTERNAL on invalid response type.
 */
FhgfsOpsErr __MessagingTk_requestResponseWithRRArgsComm(App* app,
   MessagingTk_RequestResponseArgs* rrArgs)
{
   /* note: keep in mind that there are multiple alternative response buf alloc types avilable,
      e.g. "kmalloc" or "get from store". */

   Logger* log = App_getLogger(app);
   const char* logContext = "Messaging (RPC)";

   NodeConnPool* connPool = Node_getConnPool(rrArgs->node);

   FhgfsOpsErr retVal = FhgfsOpsErr_COMMUNICATION;

   sigset_t oldSignalSet;
   fhgfs_bool useDelayedRecvInterrupt = fhgfs_false;

   unsigned bufLen; // length of shared send/recv buffer
   unsigned sendBufLen; // serialization length for sending
   ssize_t respRes = 0;
   ssize_t sendRes;

   // cleanup init
   Socket* sock = NULL;
   rrArgs->outRespBuf = NULL;
   rrArgs->outRespMsg = NULL;

   // connect
   // note: acquireStreamSocket() will fail immediately if a signal is pending

   sock = NodeConnPool_acquireStreamSocket(connPool);
   if(unlikely(!sock) )
   { // not connected
      if(!rrArgs->connEstablishFailedLogged && !Thread_isSignalPending() )
      { // only log once and only if user didn't manually interrupt with signal (to avoid log spam)
         Logger_logFormatted(log, Log_WARNING, logContext,
            "Unable to connect to: %s", Node_getNodeIDWithTypeStr(rrArgs->node) );
         Logger_logFormatted(log, Log_DEBUG, logContext,
            "Message type: %hu", NetMessage_getMsgType(rrArgs->requestMsg) );

         rrArgs->connEstablishFailedLogged = fhgfs_true;
      }
      
      return FhgfsOpsErr_COMMUNICATION;
   }

   // block interrupt signal

   if(rrArgs->useDelayedInterrupt)
   {
      SignalTk_blockSignals(fhgfs_false, &oldSignalSet); // B L O C K _ S I G s (incl. SIGINT)

      // (note: oldSignalSet is a set of previously blocked signals, so we use "!contains")
      useDelayedRecvInterrupt = !SignalTk_containsSignal(&oldSignalSet, SIGINT);
   }

   // prepare send buffer

   sendBufLen = NetMessage_getMsgLength(rrArgs->requestMsg);

   if(rrArgs->respBufType == MessagingTkBufType_BufStore)
   { // pre-alloc'ed buffer from store

      NoAllocBufferStore* bufStore = App_getMsgBufStore(app);
      bufLen = NoAllocBufferStore_getBufSize(bufStore);

      if(unlikely(bufLen < sendBufLen) )
      { // should never happen: trying to send a msg that is larger than pre-alloc'ed buf size
         Logger_logFormatted(log, Log_CRITICAL, logContext,
            "BufferStore buf size (%u) too small for msg length (%u). Message type: %hu",
            bufLen, sendBufLen, NetMessage_getMsgType(rrArgs->requestMsg) );

         retVal = FhgfsOpsErr_INTERNAL;
         goto socket_invalidate;
      }

      rrArgs->outRespBuf = NoAllocBufferStore_waitForBuf(bufStore);
   }
   else
   { // alloc'ed buffer

      bufLen = MAX(MSGTK_KMALLOC_RECV_BUF_LEN, sendBufLen);

      rrArgs->outRespBuf = (char*)os_kmalloc(bufLen);
      if(unlikely(!rrArgs->outRespBuf) )
      {
         Logger_logFormatted(log, Log_CRITICAL, logContext,
            "Buffer allocation failed. Message type: %hu; Alloc size: %u",
            NetMessage_getMsgType(rrArgs->requestMsg), bufLen);

         retVal = FhgfsOpsErr_INTERNAL;
         goto socket_invalidate;
      }
   }

   NetMessage_serialize(rrArgs->requestMsg, rrArgs->outRespBuf, sendBufLen);

   // send request

   sendRes = sock->send(sock, rrArgs->outRespBuf, sendBufLen, 0);

   if(unlikely(sendRes != (ssize_t)sendBufLen) )
      goto socket_exception;

   // receive response
   // (note: recvMsgBufEx might re-enable SIGINT if useDelayedRecvInterrupt is set)

   respRes = MessagingTk_recvMsgBufEx(app, sock, rrArgs->outRespBuf, bufLen,
      useDelayedRecvInterrupt);

   if(unlikely(respRes <= 0) )
   { // error
      if(!rrArgs->connErrLogged)
      {
         if(Thread_isSignalPending() )
            Logger_logFormatted(log, Log_NOTICE, logContext,
               "Receive interrupted by signal. Node: %s @ %s",
               Node_getNodeIDWithTypeStr(rrArgs->node), Socket_getPeername(sock) );
         else
            Logger_logFormatted(log, Log_WARNING, logContext,
               "Receive failed from %s @ %s (recv result: %lld)",
               Node_getNodeIDWithTypeStr(rrArgs->node), Socket_getPeername(sock), respRes);

         Logger_logFormatted(log, Log_DEBUG, logContext,
            "Expected response type: %u", rrArgs->respMsgType);
      }

      goto socket_invalidate;
   }

   // got response => deserialize it
   rrArgs->outRespMsg = NetMessageFactory_createFromBuf(app, rrArgs->outRespBuf, respRes);

   if(unlikely(NetMessage_getMsgType(rrArgs->outRespMsg) != rrArgs->respMsgType) )
   { // response invalid (wrong msgType)
      Logger_logErrFormatted(log, logContext,
         "Received invalid response type: %hu; expected: %d. Disconnecting: %s (%s)",
         NetMessage_getMsgType(rrArgs->outRespMsg), rrArgs->respMsgType,
         Node_getNodeIDWithTypeStr(rrArgs->node), Socket_getPeername(sock) );

      retVal = FhgfsOpsErr_INTERNAL;
      goto socket_invalidate;
   }

   // correct response => return it (through rrArgs)

   NodeConnPool_releaseStreamSocket(connPool, sock);

   if(rrArgs->useDelayedInterrupt)
      SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return FhgfsOpsErr_SUCCESS;


   // error handling (something went wrong)...

   socket_exception:
   {
      if(!rrArgs->connErrLogged)
      {
         Logger_logErrFormatted(log, logContext,
            "Communication error: Node: %s (comm result: %lld; message type: %hu)",
            Node_getNodeIDWithTypeStr(rrArgs->node),
            (long long)( (sendRes <= 0) ? sendRes : respRes),
            NetMessage_getMsgType(rrArgs->requestMsg) );
         
         rrArgs->connErrLogged = fhgfs_true;
      }
   }

   socket_invalidate:
   {
      NodeConnPool_invalidateStreamSocket(connPool, sock);
   }

   // clean up

   if(rrArgs->outRespMsg)
   {
      NetMessage_virtualDestruct(rrArgs->outRespMsg);
      rrArgs->outRespMsg = NULL;
   }

   if(rrArgs->outRespBuf)
   {
      if(rrArgs->respBufType == MessagingTkBufType_BufStore)
      {
         NoAllocBufferStore* bufStore = App_getMsgBufStore(app);
         NoAllocBufferStore_addBuf(bufStore, rrArgs->outRespBuf);
      }
      else
         os_kfree(rrArgs->outRespBuf);

      rrArgs->outRespBuf = NULL;
   }

   if(rrArgs->useDelayedInterrupt)
      SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return retVal;
}

