#ifndef MESSAGINGTK_H_
#define MESSAGINGTK_H_

#include <app/log/Logger.h>
#include <app/App.h>
#include <app/config/Config.h>
#include <common/net/message/NetMessage.h>
#include <common/net/sock/Socket.h>
#include <common/nodes/Node.h>
#include <common/threading/Thread.h>
#include <common/storage/StorageErrors.h>
#include <common/Common.h>
#include <net/message/NetMessageFactory.h>
#include <toolkit/SignalTk.h>


enum MessagingTkBufType;
typedef enum MessagingTkBufType MessagingTkBufType;

struct MessagingTk_RequestResponseArgs;
typedef struct MessagingTk_RequestResponseArgs MessagingTk_RequestResponseArgs;


extern FhgfsOpsErr __MessagingTk_requestResponseWithRRArgsComm(App* app,
   MessagingTk_RequestResponseArgs* rrArgs);

// inliners
static inline FhgfsOpsErr MessagingTk_requestResponseWithRRArgs(App* app,
   MessagingTk_RequestResponseArgs* rrArgs);
static inline FhgfsOpsErr MessagingTk_requestResponseKMalloc(App* app, Node* node,
   NetMessage* requestMsg, unsigned respMsgType, char** outRespBuf, NetMessage** outRespMsg);
static inline FhgfsOpsErr MessagingTk_requestResponse(App* app, Node* node,
   NetMessage* requestMsg, unsigned respMsgType, char** outRespBuf, NetMessage** outRespMsg);
static inline FhgfsOpsErr MessagingTk_requestResponseRetry(App* app, Node* node,
   NetMessage* requestMsg, unsigned respMsgType, char** outRespBuf, NetMessage** outRespMsg);
static inline FhgfsOpsErr MessagingTk_requestResponseRetryAutoIntr(App* app, Node* node,
   NetMessage* requestMsg, unsigned respMsgType, char** outRespBuf, NetMessage** outRespMsg);

static inline char* MessagingTk_createMsgBuf(NetMessage* msg);
static inline ssize_t MessagingTk_recvMsgBuf(App* app, Socket* sock, char* bufIn, size_t bufInLen);
static inline ssize_t MessagingTk_recvMsgBufEx(App* app, Socket* sock, char* bufIn, size_t bufInLen,
   fhgfs_bool useDelayedInterrupt);

static inline void MessagingTk_waitBeforeRetry(unsigned currentNumRetry);
static inline int MessagingTk_getRetryWaitMS(unsigned currentNumRetry);


enum MessagingTkBufType
{
   MessagingTkBufType_BufStore = 0,
   MessagingTkBufType_kmalloc  = 1, // only for small response messages (<4KB)
};

struct MessagingTk_RequestResponseArgs
{
   Node* node; // receiver
   unsigned numRetries; // 0 means infinite retries
   fhgfs_bool useDelayedInterrupt; // true to block interrupt sig for short recv wait

   NetMessage* requestMsg;
   unsigned respMsgType; // expected type of response message

   MessagingTkBufType respBufType; // defines how to get/alloc outRespBuf

   char* outRespBuf;
   NetMessage* outRespMsg;

   // internal (initialized by MessagingTk_requestResponseWithRRArgs() )
   fhgfs_bool connEstablishFailedLogged; // to avoid logging an error for each retry
   fhgfs_bool connErrLogged; // to avoid logging an error for each retry
};


/**
 * Note: rrArgs->outRespBuf must be returned/freed by the caller (depending on respBufType)
 */
FhgfsOpsErr MessagingTk_requestResponseWithRRArgs(App* app,
   MessagingTk_RequestResponseArgs* rrArgs)
{
   Logger* log = App_getLogger(app);
   const char* logContext = "Messaging (RPC)";

   fhgfs_bool infiniteRetries = rrArgs->numRetries ? fhgfs_false : fhgfs_true;

   unsigned currentRetryNum = 0;
   FhgfsOpsErr commRes;

   rrArgs->connEstablishFailedLogged = fhgfs_false;
   rrArgs->connErrLogged = fhgfs_false;

   for( ; ; ) // retry loop
   {
      App_incNumRPCs(app);

      commRes = __MessagingTk_requestResponseWithRRArgsComm(app, rrArgs);
      if(likely(commRes == FhgfsOpsErr_SUCCESS) )
         return FhgfsOpsErr_SUCCESS;
      else
      if(Thread_isSignalPending() )
      { // no retry allowed in this situation
         return FhgfsOpsErr_INTERRUPTED;
      }
      else
      if(!Node_getIsActive(rrArgs->node) )
      { // no retry allowed in this situation
         return FhgfsOpsErr_UNKNOWNNODE;
      }
      else
      if(commRes != FhgfsOpsErr_COMMUNICATION)
      { // no retry allowed in this situation
         return commRes;
      }
      if(App_getConnRetriesEnabled(app) &&
         (infiniteRetries || (currentRetryNum < rrArgs->numRetries) ) )
      { // we have a retry left
         MessagingTk_waitBeforeRetry(currentRetryNum);
         currentRetryNum++;

         if(currentRetryNum == 1) // log retry message only on first retry (to not spam the log)
         {
            Logger_logFormatted(log, Log_NOTICE, logContext,
               "Retrying communication with node: %s", Node_getNodeIDWithTypeStr(rrArgs->node) );
            Logger_logFormatted(log, Log_DEBUG, logContext,
               "Message type: %hu", NetMessage_getMsgType(rrArgs->requestMsg) );
         }
      }
      else
      { // no more retries left
         return FhgfsOpsErr_COMMUNICATION;
      }

   }
}

/**
 * Note: You probably rather want to call the alternative method, which gets buffers from the
 * store (unless you are the logger and want to avoid a deadlock wrt depleted msg buffers from the
 * store).
 * Note: Allows only a single retry. (One retry allowed because we might have gotten an already
 * broken connection from the conn pool.)
 *
 * @outRespBuf will be kmalloced and needs to be kfreed by the caller
 */
FhgfsOpsErr MessagingTk_requestResponseKMalloc(App* app, Node* node, NetMessage* requestMsg,
   unsigned respMsgType, char** outRespBuf, NetMessage** outRespMsg)
{
   MessagingTk_RequestResponseArgs rrArgs;
   FhgfsOpsErr rrRes;

   rrArgs.node = node;
   rrArgs.numRetries = 1;
   rrArgs.useDelayedInterrupt = fhgfs_false;
   rrArgs.requestMsg = requestMsg;
   rrArgs.respMsgType = respMsgType;
   rrArgs.respBufType = MessagingTkBufType_kmalloc;

   rrRes = MessagingTk_requestResponseWithRRArgs(app, &rrArgs);

   *outRespBuf = rrArgs.outRespBuf;
   *outRespMsg = rrArgs.outRespMsg;

   return rrRes;
}


/**
 * Note: Allows only a single retry. (One retry allowed because we might have gotten an already
 * broken connection from the conn pool.)
 *
 * @outRespBuf must be returned to the store - not freed!
 */
FhgfsOpsErr MessagingTk_requestResponse(App* app, Node* node, NetMessage* requestMsg,
   unsigned respMsgType, char** outRespBuf, NetMessage** outRespMsg)
{
   MessagingTk_RequestResponseArgs rrArgs;
   FhgfsOpsErr rrRes;

   rrArgs.node = node;
   rrArgs.numRetries = 1;
   rrArgs.useDelayedInterrupt = fhgfs_false;
   rrArgs.requestMsg = requestMsg;
   rrArgs.respMsgType = respMsgType;
   rrArgs.respBufType = MessagingTkBufType_BufStore;

   rrRes = MessagingTk_requestResponseWithRRArgs(app, &rrArgs);

   *outRespBuf = rrArgs.outRespBuf;
   *outRespMsg = rrArgs.outRespMsg;

   return rrRes;
}

/**
 * Note: Uses the number of retries that has been defined in the app config.
 *
 * @outRespBuf must be returned to the store - not freed!
 */
FhgfsOpsErr MessagingTk_requestResponseRetry(App* app, Node* node, NetMessage* requestMsg,
   unsigned respMsgType, char** outRespBuf, NetMessage** outRespMsg)
{
   Config* cfg = App_getConfig(app);

   MessagingTk_RequestResponseArgs rrArgs;
   FhgfsOpsErr rrRes;

   rrArgs.node = node;
   rrArgs.numRetries = Config_getConnNumCommRetries(cfg);
   rrArgs.useDelayedInterrupt = fhgfs_false;
   rrArgs.requestMsg = requestMsg;
   rrArgs.respMsgType = respMsgType;
   rrArgs.respBufType = MessagingTkBufType_BufStore;

   rrRes = MessagingTk_requestResponseWithRRArgs(app, &rrArgs);

   *outRespBuf = rrArgs.outRespBuf;
   *outRespMsg = rrArgs.outRespMsg;

   return rrRes;
}

/**
 * This version will automatically block signals including interrupt and reenables interrupt if the
 * response doesn't arrive within a few seconds.
 *
 * This is intended to help avoid cases where we miss a reply due to the user pressing ctrl+c (under
 * the assumption that the user won't care to wait a fews seconds longer for interruption).
 *
 * Note: Uses the number of retries that has been defined in the app config.
 * Note: Blocked signals will be restored to original values before this method returns.
 *
 * @outRespBuf must be returned to the store - not freed!
 */
FhgfsOpsErr MessagingTk_requestResponseRetryAutoIntr(App* app, Node* node, NetMessage* requestMsg,
   unsigned respMsgType, char** outRespBuf, NetMessage** outRespMsg)
{
   Config* cfg = App_getConfig(app);

   sigset_t oldSignalSet;

   MessagingTk_RequestResponseArgs rrArgs;
   FhgfsOpsErr rrRes;

   // keep thread interruptible for connect (will later be made uninterruptible for send/recv)
   SignalTk_blockSignals(fhgfs_true, &oldSignalSet); // B L O C K _ S I G s

   rrArgs.node = node;
   rrArgs.numRetries = Config_getConnNumCommRetries(cfg);
   rrArgs.useDelayedInterrupt = fhgfs_true;
   rrArgs.requestMsg = requestMsg;
   rrArgs.respMsgType = respMsgType;
   rrArgs.respBufType = MessagingTkBufType_BufStore;

   rrRes = MessagingTk_requestResponseWithRRArgs(app, &rrArgs);

   *outRespBuf = rrArgs.outRespBuf;
   *outRespMsg = rrArgs.outRespMsg;

   SignalTk_restoreSignals(&oldSignalSet); // U N B L O C K _ S I G s

   return rrRes;
}


/**
 * Creates a message buffer of the required size and serializes the message to it.
 * Note: Uses kmalloc, so this is only safe for messages < 128kB
 *
 * @return buffer must be kfreed by the caller
 */
char* MessagingTk_createMsgBuf(NetMessage* msg)
{
   size_t bufLen = NetMessage_getMsgLength(msg);
   char* buf = (char*)os_kmalloc(bufLen);

   NetMessage_serialize(msg, buf, buf ? bufLen : 0);

   return buf;
}

/**
 * Receive a complete message into given buffer.
 */
ssize_t MessagingTk_recvMsgBuf(App* app, Socket* sock, char* bufIn, size_t bufInLen)
{
   return MessagingTk_recvMsgBufEx(app, sock, bufIn, bufInLen, fhgfs_false);
}


/**
 * Receive a complete message into given buffer, extended version.
 *
 * Note: This is the version for pre-allocated buffers.
 *
 * @param bufIn incoming message buffer.
 * @param useDelayedInterrupt true to enable interrupt signal after a short timeout; caller must
 * ensure that the user process actually had SIGINT enabled before setting this to true.
 * @return positive message length on success, <=0 on error (e.g. -ETIMEDOUT on recv timeout,
 * -EMSGSIZE if msg too large for buffer).
 */
ssize_t MessagingTk_recvMsgBufEx(App* app, Socket* sock, char* bufIn, size_t bufInLen,
   fhgfs_bool useDelayedInterrupt)
{
   const char* logContext = "MessagingTk (recv msg)";
   Config* config = App_getConfig(app);

   const int recvNonIntrTimeoutMS = Config_getConnRecvNonIntrTimeoutMS(config);
   const int recvNormalTimeoutMS = CONN_LONG_TIMEOUT - recvNonIntrTimeoutMS; // interruptible wait

   int recvTimeoutMS = useDelayedInterrupt ? recvNonIntrTimeoutMS : recvNormalTimeoutMS;

   size_t numReceived = 0;
   ssize_t recvRes;
   size_t msgLength;

   // receive at least the message header

   recvRes = Socket_recvExactTEx(sock, bufIn, NETMSG_MIN_LENGTH, 0, recvTimeoutMS, &numReceived);

   if(unlikely(recvRes == -ETIMEDOUT) && (recvTimeoutMS == recvNonIntrTimeoutMS) )
   { // non-interruptible timeout expired => enable interrupt signal and try again
      recvTimeoutMS = recvNormalTimeoutMS;

      SignalTk_enableSigInt();

      recvRes = Socket_recvExactTEx(sock, &bufIn[numReceived], NETMSG_MIN_LENGTH, 0, recvTimeoutMS,
         &numReceived);
   }

   if(unlikely(recvRes <= 0) )
   { // socket error
      Logger* log = App_getLogger(app);
      Logger_logFormatted(log, Log_DEBUG, logContext, "Failed to receive message header from: %s",
         Socket_getPeername(sock) );

      goto socket_exception;
   }

   msgLength = NetMessage_extractMsgLengthFromBuf(bufIn);

   if(msgLength <= numReceived)
      return msgLength; // success (msg had no additional payload)


   // receive the message payload part

   if(unlikely(msgLength > bufInLen) )
   { // message too big to be accepted
      Logger* log = App_getLogger(app);
      Logger_logFormatted(log, Log_WARNING, logContext,
         "Received a message that is too large from: %s (bufLen: %lld, msgLen: %lld)",
         Socket_getPeername(sock), (long long)bufInLen, (long long)msgLength);

      return -EMSGSIZE;
   }

   recvRes = Socket_recvExactTEx(sock, &bufIn[numReceived], msgLength-numReceived, 0, recvTimeoutMS,
      &numReceived);

   if(unlikely(recvRes == -ETIMEDOUT) && (recvTimeoutMS == recvNonIntrTimeoutMS) )
   { // non-interruptible timeout expired => enable interrut signal and try again
      recvTimeoutMS = recvNormalTimeoutMS;

      SignalTk_enableSigInt();

      recvRes = Socket_recvExactTEx(sock, &bufIn[numReceived], NETMSG_MIN_LENGTH, 0, recvTimeoutMS,
         &numReceived);
   }

   if(unlikely(recvRes <= 0) )
      goto socket_exception;

   // success
   return msgLength;


   socket_exception:
   {
      Logger* log = App_getLogger(app);

      if(Thread_isSignalPending() )
         Logger_logFormatted(log, Log_DEBUG, logContext,
            "Receive interrupted by signal",
            Socket_getPeername(sock), (long long)recvRes);
      else
         Logger_logFormatted(log, Log_DEBUG, logContext,
            "Receive failed from: %s (ErrCode: %lld)",
            Socket_getPeername(sock), (long long)recvRes);
   }

   return recvRes;
}

void MessagingTk_waitBeforeRetry(unsigned currentNumRetry)
{
   int retryWaitMS = MessagingTk_getRetryWaitMS(currentNumRetry);

   if(retryWaitMS)
      Thread_sleep(retryWaitMS);
}

int MessagingTk_getRetryWaitMS(unsigned currentNumRetry)
{
   // note: keep in sync with __Config_initConnNumCommRetries()

   int retryWaitMS ;

   if(!currentNumRetry)
   {
      retryWaitMS = 0;
   }
   else
   if(currentNumRetry <= 12) // 1st minute
   {
      retryWaitMS = 5000;
   }
   else
   if(currentNumRetry <= 24) // 2nd to 5th minute
   {
      retryWaitMS = 20000;
   }
   else // after 5th minute
   {
      retryWaitMS = 60000;
   }

   return retryWaitMS;
}

#endif /*MESSAGINGTK_H_*/
