#include <app/log/Logger.h>
#include <app/App.h>
#include <app/config/Config.h>
#include <common/toolkit/vector/StrCpyVec.h>
#include <common/toolkit/vector/IntCpyVec.h>
#include <common/toolkit/LockingTk.h>
#include <common/storage/StorageErrors.h>
#include <common/toolkit/StringTk.h>
#include <filesystem/ProcFs.h>
#include <os/OsCompat.h>
#include <os/OsTypeConversion.h>
#include "FhgfsOpsHelper.h"
#include "FhgfsOpsFile.h"
#include "FhgfsOpsDir.h"
#include "FhgfsOpsInode.h"
#include "FhgfsOpsIoctl.h"
#include "FhgfsOpsSuper.h"
#include "FhgfsOps_versions.h"
#include "FhgfsOpsPages.h"

#include <linux/aio.h>
#include <linux/writeback.h>
#include <linux/mm.h>
#include <linux/mpage.h>
#include <linux/backing-dev.h>
#include <linux/pagemap.h>


#ifdef CONFIG_COMPAT
#include <asm/compat.h>
#endif // CONFIG_COMPAT

#define FHGFSOPS_WRITEPAGES_POINTER_RESERVE   (1024) /* number of page pointers in write buffer */

#define INITIAL_FIND_PAGES (256) // search initially for this number of pages

/**
 * Operations for files with cache type "buffered" and "none".
 */
struct file_operations fhgfs_file_buffered_ops =
{
   .open             = FhgfsOps_open,
   .release          = FhgfsOps_release,
   .read             = FhgfsOps_read,
   .write            = FhgfsOps_write,
   .fsync            = FhgfsOps_fsync,
   .flush            = FhgfsOps_flush,
   .llseek           = FhgfsOps_llseek,
   .flock            = FhgfsOps_flock,
   .lock             = FhgfsOps_lock,
   .mmap             = FhgfsOps_mmap,
   .unlocked_ioctl   = FhgfsOpsIoctl_ioctl,
#ifdef CONFIG_COMPAT
   .compat_ioctl     = FhgfsOpsIoctl_compatIoctl,
#endif // CONFIG_COMPAT

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,16,0)
   .splice_read  = generic_file_splice_read,
   .splice_write = iter_file_splice_write,
#else
   .splice_read  = generic_file_splice_read,
   .splice_write = generic_file_splice_write,
#endif // LINUX_VERSION_CODE

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,22)
   .sendfile   = generic_file_sendfile, // removed in 2.6.23 (now handled via splice)
#endif // LINUX_VERSION_CODE
};

/**
 * Operations for files with cache type "paged".
 */
struct file_operations fhgfs_file_pagecache_ops =
{
   .open                = FhgfsOps_open,
   .release             = FhgfsOps_release,
   .read                = FhgfsOps_readPaged,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,16,0)
   .write               = new_sync_write, // "new" here means iter-based (so it calls write_iter() )
   .read_iter           = FhgfsOps_read_iter, // replacement for aio_read
   .write_iter          = FhgfsOps_write_iter, // replacement for aio_write
#else
   .write               = do_sync_write, // (calls aio_write() )
   .aio_read            = FhgfsOps_aio_read,
   .aio_write           = FhgfsOps_aio_write,
#endif // LINUX_VERSION_CODE
   .fsync               = FhgfsOps_fsync,
   .flush               = FhgfsOps_flush,
   .llseek              = FhgfsOps_llseek,
   .flock               = FhgfsOps_flock,
   .lock                = FhgfsOps_lock,
   .mmap                = FhgfsOps_mmap,
   .unlocked_ioctl      = FhgfsOpsIoctl_ioctl,
#ifdef CONFIG_COMPAT
   .compat_ioctl        = FhgfsOpsIoctl_compatIoctl,
#endif // CONFIG_COMPAT

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,16,0)
   .splice_read  = generic_file_splice_read,
   .splice_write = iter_file_splice_write,
#else
   .splice_read  = generic_file_splice_read,
   .splice_write = generic_file_splice_write,
#endif // LINUX_VERSION_CODE

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,22)
   .sendfile   = generic_file_sendfile, // removed in 2.6.23 (now handled via splice)
#endif // LINUX_VERSION_CODE
};

/**
 * Operations for files with cache type "hybrid" (i.e. read paged, write buffered).
 */
struct file_operations fhgfs_file_hybridcache_ops =
{
   .open                = FhgfsOps_open,
   .release             = FhgfsOps_release,
   .read                = FhgfsOps_readPagedHybrid,
   .write               = FhgfsOps_writeHybrid,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,16,0)
   .read_iter           = FhgfsOps_read_iter, // replacement for aio_read
#else
   .aio_read            = FhgfsOps_aio_read,
#endif // LINUX_VERSION_CODE
   .fsync               = FhgfsOps_fsync,
   .flush               = FhgfsOps_flush,
   .llseek              = FhgfsOps_llseek,
   .flock               = FhgfsOps_flock,
   .lock                = FhgfsOps_lock,
   .mmap                = FhgfsOps_mmap,
   .unlocked_ioctl      = FhgfsOpsIoctl_ioctl,
#ifdef CONFIG_COMPAT
   .compat_ioctl        = FhgfsOpsIoctl_compatIoctl,
#endif // CONFIG_COMPAT

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,16,0)
   .splice_read  = generic_file_splice_read,
   .splice_write = iter_file_splice_write,
#else
   .splice_read  = generic_file_splice_read,
   .splice_write = generic_file_splice_write,
#endif // LINUX_VERSION_CODE

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,22)
   .sendfile   = generic_file_sendfile, // removed in 2.6.23 (now handled via splice)
#endif // LINUX_VERSION_CODE
};

struct file_operations fhgfs_dir_ops =
{
   .open             = FhgfsOps_opendirIncremental,
   .release          = FhgfsOps_releasedir,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)
   .iterate       = FhgfsOps_iterateIncremental, // linux 3.11 renamed readdir to iterate
#else
   .readdir       = FhgfsOps_readdirIncremental, // linux 3.11 renamed readdir to iterate
#endif // LINUX_VERSION_CODE
   .read             = generic_read_dir, // just returns the appropriate error code
   .fsync            = FhgfsOps_fsync,
   .llseek           = FhgfsOps_llseekdir,
   .unlocked_ioctl   = FhgfsOpsIoctl_ioctl,
#ifdef CONFIG_COMPAT
   .compat_ioctl     = FhgfsOpsIoctl_compatIoctl,
#endif // CONFIG_COMPAT
};

/**
 * Operations for files with cache type "buffered" and "none".
 */
struct address_space_operations fhgfs_address_ops =
{
   .readpage       = FhgfsOpsPages_readpage,
   .readpages      = FhgfsOpsPages_readpages,
   .writepage      = FhgfsOpsPages_writepage,
#ifndef KERNEL_HAS_NO_WRITE_CACHE_PAGES
   .writepages     = FhgfsOpsPages_writepages,
#endif // KERNEL_HAS_NO_WRITE_CACHE_PAGES
   .set_page_dirty = __set_page_dirty_nobuffers,
   .direct_IO      = FhgfsOps_directIO,

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
   .prepare_write = FhgfsOps_prepare_write,
   .commit_write  = FhgfsOps_commit_write,
#else
   .write_begin   = FhgfsOps_write_begin,
   .write_end     = FhgfsOps_write_end,
#endif // LINUX_VERSION_CODE
};

/**
 * Operations for files with cache type "paged".
 */
struct address_space_operations fhgfs_address_pagecache_ops =
{
   .readpage       = FhgfsOpsPages_readpage,
   .readpages      = FhgfsOpsPages_readpages,
   .writepage      = FhgfsOpsPages_writepage,
#ifndef KERNEL_HAS_NO_WRITE_CACHE_PAGES
   .writepages     = FhgfsOpsPages_writepages,
#endif
   .set_page_dirty = __set_page_dirty_nobuffers,
   .direct_IO      = FhgfsOps_directIO,

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)
   .prepare_write = FhgfsOps_prepare_write,
   .commit_write  = FhgfsOps_commit_write,
#else
   .write_begin   = FhgfsOps_write_begin,
   .write_end     = FhgfsOps_write_end,
#endif // LINUX_VERSION_CODE
};


/**
 * @param origin dirs allow only SEEK_SET (via seekdir/rewinddir from userspace).
 */
loff_t FhgfsOps_llseekdir(struct file *file, loff_t offset, int origin)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_llseekDir";
   struct inode* inode = file_inode(file);

   loff_t retVal = 0;
   FsDirInfo* dirInfo = __FhgfsOps_getDirInfo(file);

   if(unlikely(Logger_getLogLevel(log) >= 5) )
      FhgfsOpsHelper_logOp(5, app, file->f_dentry, inode, logContext);

   if(origin != SEEK_SET)
      return -EINVAL;

   retVal = generic_file_llseek_unlocked(file, offset, origin);
   if(likely(retVal >= 0) )
   {
      // invalidate any retrieved contents to keep things in sync with server offset
      StrCpyVec* contents = FsDirInfo_getDirContents(dirInfo);
      uint64_t incrementalOffset = (offset < 2) ? 0 : (offset - 2); // "-2" because we use the
         // first two entries for "." and ".."

      StrCpyVec_clear(contents);
      FsDirInfo_setCurrentContentsPos(dirInfo, 0);

      FsDirInfo_setServerOffset(dirInfo, -1);
      FsDirInfo_setIncrementalOffset(dirInfo, incrementalOffset);
      FsDirInfo_setEndOfDir(dirInfo, fhgfs_false);
   }

   return retVal;
}

loff_t FhgfsOps_llseek(struct file *file, loff_t offset, int origin)
{
   const char* logContext = "FhgfsOps_llseek";
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);

   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   fhgfs_bool isGloballyLockedAppend =
      FsFileInfo_getAppending(fileInfo) && Config_getTuneUseGlobalAppendLocks(cfg);

   loff_t retVal = 0;
   struct inode *inode = file->f_mapping->host;

   FhgfsIsizeHints iSizeHints;

   if(unlikely(Logger_getLogLevel(log) >= Log_SPAM) )
   {
      FhgfsOpsHelper_logOpMsg(Log_SPAM, app, file->f_dentry, inode, logContext,
         "offset: %ld directive: %d", offset, origin);
   }

   /* note: globally locked append with SEEK_CUR is a special case, because we need to flush
      the cache to find out the current offset (which is not required without append) */
   if( (origin == SEEK_END) ||
       (isGloballyLockedAppend && (origin == SEEK_CUR) ) )
   { // seek to position relative to end-of-file => flush cache and update current file size first

      // (note: refreshInode() also flushes caches for correct file size)

      retVal = __FhgfsOps_refreshInode(app, inode, NULL, &iSizeHints);
      if(retVal)
         goto clean_up;

      spin_lock(&inode->i_lock); // L O C K

      // SEEK_CUR reads (and modifies) f_pos, so in buffered append mode move to end first
      if(origin == SEEK_CUR)
         file->f_pos = inode->i_size;

      retVal = generic_file_llseek_unlocked(file, offset, origin);

      spin_unlock(&inode->i_lock); // U N L O C K
   }
   else
   { // abolute or relative-to-current_pos seeks => generic stuff
      retVal = generic_file_llseek_unlocked(file, offset, origin);
   }


clean_up:
   // clean-up

   FhgfsOpsHelper_logOpDebug(app, file->f_dentry, inode, logContext, "retVal: %lld",
      retVal);

   return retVal;
}

/**
 * Note: Currently unsused method, as we're using the kernel's generic_readlink function.
 */
int FhgfsOps_readlink(struct dentry* dentry, char __user* buf, int size)
{
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_readlink";

   int retVal;
   struct inode* inode = dentry->d_inode;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   if(unlikely(Logger_getLogLevel(log) >= 5) )
      FhgfsOpsHelper_logOp(5, app, dentry, inode, logContext);

   // check user buffer
   if(unlikely(!access_ok(VERIFY_WRITE, buf, size) ) )
      return -EFAULT;

   FhgfsInode_entryInfoReadLock(fhgfsInode); // LOCK EntryInfo

   retVal = FhgfsOpsHelper_readlink(app, FhgfsInode_getEntryInfo(fhgfsInode), buf, size);

   FhgfsInode_entryInfoReadUnlock(fhgfsInode); // UNLOCK EntryInfo


   return retVal;
}

/**
 * Opens a directory and prepares the handle for incremental readdir().
 */
int FhgfsOps_opendirIncremental(struct inode* inode, struct file* file)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_opendirIncremental";

   int retVal = 0;
   //struct dentry* dentry = file->f_dentry;
   FsDirInfo* dirInfo;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
      FhgfsOpsHelper_logOp(5, app, file->f_dentry, inode, logContext);

   //retVal = __FhgfsOps_refreshInode(app, inode); // not necessary
   if(!retVal)
   { // success
      dirInfo = FsDirInfo_construct(app);
      __FhgfsOps_setDirInfo(dirInfo, file);
   }

   return retVal;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)
int FhgfsOps_iterateIncremental(struct file* file, struct dir_context* ctx)
#else
int FhgfsOps_readdirIncremental(struct file* file, void* buf, filldir_t filldir)
#endif // LINUX_VERSION_CODE
{
   // note: we assume that dir reading is strictly sequential, thus we don't actually use f_pos
      // here (and instead keep a separate server offset var, which is not sequential). if the user
      // seeks, llseekDir will invalidate any retrieved contents and set the new offsets in the
      // dirinfo object, so that we can continue our virtually sequential access even in this case.
      // however, the user just sees a sequential offset, so we use f_pos to manage that.

   struct super_block* superBlock = file->f_dentry->d_sb;
   App* app = FhgfsOps_getApp(superBlock);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_readdirIncremental";

   int retVal = 0;
   FsDirInfo* dirInfo = __FhgfsOps_getDirInfo(file);
   struct dentry* dentry = file->f_dentry;
   struct inode* inode = file_inode(file);
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   loff_t* pos; // position/offset


   if(unlikely(Logger_getLogLevel(log) >= 5) )
      FhgfsOpsHelper_logOp(5, app, file->f_dentry, inode, logContext);


   #if LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)
      pos = &(ctx->pos);
   #else
      pos = &(file->f_pos);
   #endif // LINUX_VERSION_CODE


   FhgfsInode_entryInfoReadLock(fhgfsInode); // LOCK EntryInfo

   //log.log(5, std::string("dirContents.size: ") + StringTk::intToStr(dirContents->size() ) );;
   //log.log(5, std::string("offset: ") + StringTk::intToStr(offset) );;

   switch(*pos)
   {
      case 0:
      {
         u64 ino = dentry->d_inode->i_ino;

         if (is_32bit_api() && ino > UINT_MAX)
         {  // 32-bit applications cannot handle 64bit inodes and would fail with EOVERFLOW
            ino = ino >> 32;
         }

         #if LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)
            if(!dir_emit(ctx, ".", 1, ino, DT_DIR) )
               break;
         #else
            if(filldir(buf, ".", 1, *pos, ino, DT_DIR) < 0)
               break;
         #endif // LINUX_VERSION_CODE

         (*pos)++;

         LOG_DEBUG_FORMATTED(log, 5, logContext, "filled[%lld]: %s",
            (long long)0, ".");
      }
         /* fall through */
      case 1:
      {
         u64 inoParent = parent_ino(dentry);

         if (is_32bit_api() && inoParent > UINT_MAX)
         {  // 32-bit applications cannot handle 64bit inodes and would fail with EOVERFLOW
            inoParent = inoParent >> 32;
         }

         #if LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)
            if(!dir_emit(ctx, "..", 2, inoParent, DT_DIR) )
               break;
         #else
            if(filldir(buf, "..", 2, *pos, inoParent, DT_DIR) < 0)
               break;
         #endif // LINUX_VERSION_CODE

         (*pos)++;

         LOG_DEBUG_FORMATTED(log, 5, logContext, "filled[%lld]: %s",
            (long long)1, "..");
      }
         /* fall through */
      default:
      {
         size_t contentsPos;
         StrCpyVec* dirContents = FsDirInfo_getDirContents(dirInfo);
         IntCpyVec* dirContentsTypes = FsDirInfo_getDirContentsTypes(dirInfo);
         StrCpyVec* dirContentIDs = FsDirInfo_getEntryIDs(dirInfo);
         size_t contentsLength;

         for( ; ; ) // loop as long as filldir callback swallows more entries
         {
            int refreshRes;
            char* currentName;
            DirEntryType currentEntryType;
            unsigned currentOSEntryType;
            const char* currentEntryID;
            uint64_t currentIno;

            refreshRes = FhgfsOpsHelper_refreshDirInfoIncremental(app,
               FhgfsInode_getEntryInfo(fhgfsInode), dirInfo, fhgfs_false);
            if(unlikely(refreshRes) )
            { // error occurred
               retVal = refreshRes;
               break;
            }

            contentsPos = FsDirInfo_getCurrentContentsPos(dirInfo);
            contentsLength = StrCpyVec_length(dirContents);

            LOG_DEBUG_FORMATTED(log, 5, logContext,
               "pos: %lld, contentsPos: %lld/%lld, endOfDir: %s",
               (long long)*pos, (long long)contentsPos,
               (long long)contentsLength, FsDirInfo_getEndOfDir(dirInfo) ? "yes" : "no");


            // refreshDirInfoInc guarantees that we either have a valid range for current offset
            // or that contentsLength is empty
            if(!contentsLength)
            { // end of dir
               LOG_DEBUG(log, 5, logContext, "reached end of dir");
               break;
            }

            currentName = StrCpyVec_at(dirContents, contentsPos);

            currentEntryType = IntCpyVec_at(dirContentsTypes, contentsPos);
            currentOSEntryType = OsTypeConv_dirEntryTypeToOS(currentEntryType);

            currentEntryID = StrCpyVec_at(dirContentIDs, contentsPos);
            currentIno = FhgfsInode_generateInodeID(superBlock, currentEntryID,
               strlen(currentEntryID) );
            if (is_32bit_api() && currentIno > UINT_MAX)
            {  // 32-bit applications cannot handle 64bit inodes and would fail with EOVERFLOW
               currentIno =  currentIno >> 32;
            }


            #if LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)
               if(!dir_emit(
                  ctx, currentName, strlen(currentName), currentIno, currentOSEntryType) )
                  break;
            #else
               if(filldir(
                  buf, currentName, strlen(currentName), *pos, currentIno, currentOSEntryType) < 0)
                  break;
            #endif // LINUX_VERSION_CODE


            LOG_DEBUG_FORMATTED(log, 5, logContext, "filled: %s", currentName);

            (*pos)++;
            FsDirInfo_setCurrentContentsPos(dirInfo, contentsPos+1);

         } // end of for-loop
      } break;
   } // end of switch

   // clean-up
   FhgfsInode_entryInfoReadUnlock(fhgfsInode); // UNLOCK EntryInfo

   return retVal;
}

/**
 * Note: This works for _opendir() and for _opendirIncremental().
 */
int FhgfsOps_releasedir(struct inode* inode, struct file* file)
{
   const char* logContext = "FhgfsOps_releasedir";

   FsObjectInfo* fsObjectInfo = __FhgfsOps_getObjectInfo(file);

   App* app = FsObjectInfo_getApp(fsObjectInfo);

   FhgfsOpsHelper_logOp(Log_SPAM, app, file->f_dentry, inode, logContext);

   FsObjectInfo_virtualDestruct(fsObjectInfo);

   return 0;
}

/**
 * Open a file, may be called from vfs or lookup/atomic open.
 *
 * @param lookupInfo is NULL if this is a direct open call from the vfs
 */
int FhgfsOps_openReferenceHandle(App* app, struct inode* inode, struct file* file,
   unsigned openFlags, LookupIntentInfoOut* lookupInfo)
{
   Config* cfg = App_getConfig(app);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_openReferenceHandle";

   struct super_block* sb = inode->i_sb;
   struct dentry* dentry = file->f_dentry;

   int retVal = 0;
   int fhgfsOpenFlags;
   FileHandleType handleType;
   FhgfsOpsErr openRes;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   if(unlikely(Logger_getLogLevel(log) >= 4) )
      FhgfsOpsHelper_logOp(Log_DEBUG, app, dentry, inode, logContext);

   fhgfsOpenFlags = OsTypeConv_openFlagsOsToFhgfs(openFlags, __FhgfsOps_isPagedMode(sb) );

   openRes = FhgfsInode_referenceHandle(fhgfsInode, fhgfsOpenFlags, fhgfs_false, lookupInfo,
      &handleType);

   LOG_DEBUG_FORMATTED(log, 4, logContext, "remoting complete. result: %s",
      FhgfsOpsErr_toErrString(openRes) );

   if(openRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(openRes);
   }
   else
   { // success => file is open (=> handle open flags)
      FsFileInfo* fileInfo = FsFileInfo_construct(app, fhgfsOpenFlags, handleType);

      // handle O_APPEND
      if(file->f_flags & O_APPEND)
         FsFileInfo_setAppending(fileInfo, fhgfs_true);

      // handle O_DIRECT + disabled caching
      if( (file->f_flags & O_DIRECT) ||
          ( (file->f_flags & O_APPEND) && !Config_getTuneUseBufferedAppend(cfg) ) ||
          (Config_getTuneFileCacheTypeNum(cfg) == FILECACHETYPE_None) )
      { // disable caching
         FsFileInfo_setAllowCaching(fileInfo, fhgfs_false);
      }

      __FhgfsOps_setFileInfo(fileInfo, file);
   }

   return retVal;
}

/**
 * Open a file, vfs interface
 */
int FhgfsOps_open(struct inode* inode, struct file* file)
{
   const char* logContext = "FhgfsOps_open";

   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);

   struct dentry* dentry = file->f_dentry;

   unsigned openFlags = file->f_flags;
   LookupIntentInfoOut* lookupInfo = NULL; // not available for direct open

   if(unlikely(Logger_getLogLevel(log) >= 4) )
      FhgfsOpsHelper_logOp(4, app, dentry, inode, logContext);

   return FhgfsOps_openReferenceHandle(app, inode, file, openFlags, lookupInfo);
}

/**
 * Close a file.
 *
 * Note: We only got one shot, even in case of an error.
 */
int FhgfsOps_release(struct inode* inode, struct file* file)
{
   const char* logContext = "FhgfsOps_release";

   int retVal = 0;
   FhgfsOpsErr closeRes;
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   FsObjectInfo* fsObjectInfo = __FhgfsOps_getObjectInfo(file);
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   FileHandleType handleType = FsFileInfo_getHandleType(fileInfo);

   App* app = FsObjectInfo_getApp(fsObjectInfo);

   FhgfsOpsHelper_logOp(Log_SPAM, app, file->f_dentry, inode, logContext);

   if(unlikely(!fileInfo) )
   { // invalid file handle
      return -EBADF;
   }

   FhgfsOps_releaseCancelLocks(inode, file); // cancel all locks that were not properly released yet

   closeRes = FhgfsInode_releaseHandle(fhgfsInode, handleType);

   if(closeRes != FhgfsOpsErr_SUCCESS)
   { // error
      retVal = FhgfsOpsErr_toSysErr(closeRes);
   }

   // note: we free the fileInfo no matter whether the communication succeeded or not
   //    (because _release() won't be called again even if it didn't succeed)

   FsObjectInfo_virtualDestruct( (FsObjectInfo*)fileInfo);
   __FhgfsOps_setFileInfo( (FsFileInfo*)NULL, file);

   FhgfsInode_invalidateCache(fhgfsInode);

   // warning: linux vfs won't return this result to user apps. only flush() res is passed to apps.
   return retVal;
}

/**
 * Called during file close to unlock remaining entry locks and range locks that were not properly
 * unlocked by the user-space application yet.
 */
int FhgfsOps_releaseCancelLocks(struct inode* inode, struct file* file)
{
   int retVal = 0;
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   RemotingIOInfo ioInfo;
   FhgfsOpsErr unlockRes;

   /* (note: it is very unlikely that an application will use entry and range locking together on
      the same file, so we have no special optimization regarding EntryMinInfoCopy for that case) */

   if(FsFileInfo_getUsedEntryLocking(fileInfo) )
   { // entry locks were used with this file handle
      int64_t clientFD = __FhgfsOps_getCurrentLockFD(file);

      FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

      FhgfsInode_entryInfoReadLock(fhgfsInode); // LOCK EntryInfo

      unlockRes = FhgfsOpsHelper_unlockEntryWithAsyncRetry(FhgfsInode_getEntryInfo(fhgfsInode), &ioInfo,
         clientFD);

      FhgfsInode_entryInfoReadUnlock(fhgfsInode); // UNLOCK EntryInfo

      if(!retVal)
         retVal = FhgfsOpsErr_toSysErr(unlockRes);
   }

   /* (note: FhgfsInode_getNumRangeLockPIDs() below is a shortcut to save the time for mutex locking
      if no range locks were used for this inode.) */

   if(FhgfsInode_getNumRangeLockPIDs(fhgfsInode) &&
      FhgfsInode_removeRangeLockPID(fhgfsInode, __FhgfsOps_getCurrentLockPID() ) )
   { // current pid used range locking on this inode
      int ownerPID = __FhgfsOps_getCurrentLockPID();

      FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

      FhgfsInode_entryInfoReadLock(fhgfsInode); // LOCK EntryInfo

      unlockRes = FhgfsOpsHelper_unlockRangeWithAsyncRetry(FhgfsInode_getEntryInfo(fhgfsInode), &ioInfo,
         ownerPID);

      FhgfsInode_entryInfoReadUnlock(fhgfsInode); // UNLOCK EntryInfo

      if(!retVal)
         retVal = FhgfsOpsErr_toSysErr(unlockRes);
   }

   return retVal;
}

/**
 * Called by flock syscall.
 *
 * @return 0 on success, negative linux error code otherwise
 */
int FhgfsOps_flock(struct file* file, int cmd, struct file_lock* fileLock)
{
   const char* logContext = __func__;

   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);

   struct inode* inode = file_inode(file);
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   fhgfs_bool useGlobalFileLocks = Config_getTuneUseGlobalFileLocks(cfg);
   FhgfsOpsErr globalLockRes = FhgfsOpsErr_SUCCESS;
   int lockTypeFlags;

   lockTypeFlags = OsTypeConv_flockTypeToFhgfs(fileLock);

   if(unlikely(Logger_getLogLevel(log) >= Log_SPAM) )
      FhgfsOpsHelper_logOpMsg(Log_SPAM, app, file->f_dentry, inode, logContext, "lockType: %s",
         LockingTk_lockTypeToStr(lockTypeFlags) );


   // flush buffers before removing a global lock

   if(useGlobalFileLocks && (lockTypeFlags & ENTRYLOCKTYPE_LOCKOPS_REMOVE) )
   {
      int flushRes = __FhgfsOps_flush(app, file, fhgfs_false, fhgfs_false, fhgfs_true,
         fhgfs_false);

      /* note: can't return error here and must continue, because local unlock must always be done
         to avoid BUG() statement being triggered in locks_remove_flock() on cleanup after kill */
      if(unlikely(flushRes < 0) )
         Logger_logFormatted(log, Log_NOTICE, logContext,
            "Flushing before unlock failed. Continuing anyways. flushRes: %d", flushRes);
   }

   // global locking

   if(useGlobalFileLocks)
   {
      FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
      RemotingIOInfo ioInfo;

      FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

      FsFileInfo_setUsedEntryLocking(fileInfo);

      FhgfsInode_entryInfoReadLock(fhgfsInode); // LOCK EntryInfo

      globalLockRes = FhgfsOpsRemoting_flockEntry(FhgfsInode_getEntryInfo(fhgfsInode), &ioInfo,
         (size_t)fileLock->fl_file, fileLock->fl_pid, lockTypeFlags);

      FhgfsInode_entryInfoReadUnlock(fhgfsInode); // UNLOCK EntryInfo

      LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "remoting complete. result: %s",
         FhgfsOpsErr_toErrString(globalLockRes) );

      /* note: local unlock must always be done for cleanup (otherwise e.g. killing a process
         holding a lock results in the BUG() statement being triggered in  locks_remove_flock() ) */
      if( (globalLockRes != FhgfsOpsErr_SUCCESS) &&
         !(lockTypeFlags & ENTRYLOCKTYPE_LOCKOPS_REMOVE) )
         return FhgfsOpsErr_toSysErr(globalLockRes);
   }

   // local locking

   {
      int localLockRes = flock_lock_file_wait(file, fileLock);
      if(!useGlobalFileLocks)
         return localLockRes;

      if(localLockRes &&
         (globalLockRes == FhgfsOpsErr_SUCCESS) &&
         (lockTypeFlags & ENTRYLOCKTYPE_LOCKOPS_ADD) )
      { // sanity check
         Logger_logFormatted(log, Log_NOTICE, logContext,
            "Unexpected: Global locking succeeded, but local locking failed. SysErr: %d",
            localLockRes);
      }
   }

   // flush buffers after we got a new global lock

   if(useGlobalFileLocks &&
      (lockTypeFlags & ENTRYLOCKTYPE_LOCKOPS_ADD) &&
      (globalLockRes == FhgfsOpsErr_SUCCESS) )
   {
      int flushRes = __FhgfsOps_flush(app, file, fhgfs_false, fhgfs_false, fhgfs_true,
         fhgfs_false);

      if(unlikely(flushRes < 0) )
         return flushRes; // flush error occured
   }

   return FhgfsOpsErr_toSysErr(globalLockRes);
}

/**
 * Called by fcntl syscall (F_GETLK, F_SETLK, F_SETLKW) for file range locking.
 *
 * @return 0 on success, negative linux error code otherwise
 */
int FhgfsOps_lock(struct file* file, int cmd, struct file_lock* fileLock)
{
   const char* logContext = "FhgfsOps_lock (fcntl)";

   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);

   struct inode* inode = file_inode(file);
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   fhgfs_bool useGlobalFileLocks = Config_getTuneUseGlobalFileLocks(cfg);
   FhgfsOpsErr globalLockRes = FhgfsOpsErr_SUCCESS;
   int lockTypeFlags;


   // handle user request for conflicting locks (always local-only currently, see notes below)

   if(cmd == F_GETLK)
   { // get confliciting lock

      /* note: it's questionable if returning remote locks makes sense (because the local app could
         misinterpret the pid of the lock-holder), so we do local only for now. */

      #ifdef KERNEL_HAS_TEST_LOCK_VOID_TWOARGS

         posix_test_lock(file, fileLock);

         /* note: "fileLock->fl_type != F_UNLCK" would tell us now whether a conflicting local lock
            was found */
      #else
         /* note: there are a lots of different versions of posix_test_lock() in the
            different kernels, so we only support the most recent one for now (kernels >=2.6.23).
            later, we can switch to remote conflicts check and thus won't need the posix_test_lock
            calls, anyways. */

         fileLock->fl_type = F_UNLCK;

      #endif // KERNEL_HAS_TEST_LOCK_VOID_TWOARGS


      return 0;
   }


   lockTypeFlags = OsTypeConv_flockTypeToFhgfs(fileLock);

   if(unlikely(Logger_getLogLevel(log) >= Log_SPAM) )
      FhgfsOpsHelper_logOpMsg(Log_SPAM, app, file->f_dentry, inode, logContext,
         "lockType: %s; start: %lld; end: %lld", LockingTk_lockTypeToStr(lockTypeFlags),
         (long long)fileLock->fl_start, (long long)fileLock->fl_end);


   // flush buffers before removing a global lock

   if(useGlobalFileLocks &&
      (lockTypeFlags & ENTRYLOCKTYPE_LOCKOPS_REMOVE) )
   {
      int flushRes = __FhgfsOps_flush(app, file, fhgfs_false, fhgfs_false, fhgfs_true, fhgfs_false);

      /* note: can't return error here and must continue, because local unlock must always be done
         to avoid BUG() statement being triggered in locks_remove_flock() on cleanup after kill */
      if(unlikely(flushRes < 0) )
         Logger_logFormatted(log, Log_NOTICE, logContext,
            "Flushing before unlock failed. Continuing anyways. flushRes: %d", flushRes);
   }

   // global locking

   if(useGlobalFileLocks)
   {
      FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
      RemotingIOInfo ioInfo;

      FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

      FhgfsInode_addRangeLockPID(fhgfsInode, fileLock->fl_pid);

      FhgfsInode_entryInfoReadLock(fhgfsInode); // LOCK EntryInfo

      globalLockRes = FhgfsOpsRemoting_flockRange(FhgfsInode_getEntryInfo(fhgfsInode), &ioInfo,
         fileLock->fl_pid, lockTypeFlags, fileLock->fl_start, fileLock->fl_end);

      FhgfsInode_entryInfoReadUnlock(fhgfsInode); // UNLOCK EntryInfo

      LOG_DEBUG_FORMATTED(log, Log_DEBUG, logContext, "remoting complete. result: %s",
         FhgfsOpsErr_toErrString(globalLockRes) );

      /* note: local unlock must always be done for cleanup (otherwise e.g. killing a process
         holding a lock results in the BUG() statement being triggered in  locks_remove_flock() ) */
      if( (globalLockRes != FhgfsOpsErr_SUCCESS) &&
         !(lockTypeFlags & ENTRYLOCKTYPE_LOCKOPS_REMOVE) )
         return FhgfsOpsErr_toSysErr(globalLockRes);
   }

   // local locking

   {
      /* note on local+global locking:
         we need to call posix_lock_file_wait() even with global locks, because inode->i_flock needs
         to be set, so that locks_remove_posix() gets active (via filp_close() ) and thus the
         condition "Record locks are not inherited by a child created via fork(2), but are preserved
         across an execve(2)." [man 2 fcntl] holds.
         Otherwise we wouldn't be notified about an unlock on parent process exit, as there are
         still references to the filp and thus our ->release() isn't invoked. (See trac #271) */

      /* note on local/global locking order:
         local locking needs to be done after global locking, because otherwise if global locking
         failed we wouldn't know how to undo the local locking (e.g. if the process acquires a
         shared lock for the second time or does a merge with existing ranges). */

      int localLockRes = posix_lock_file_wait(file, fileLock);

      //printk_fhgfs_debug(KERN_WARNING, "posix_lock_file result=%d, cmd=%d\n", localLockRes, cmd);

      if(!useGlobalFileLocks)
         return localLockRes;

      if(localLockRes &&
         (globalLockRes == FhgfsOpsErr_SUCCESS) &&
         (lockTypeFlags & ENTRYLOCKTYPE_LOCKOPS_ADD) )
      { // sanity check
         Logger_logFormatted(log, Log_NOTICE, logContext,
            "Unexpected: Global locking succeeded, but local locking failed. SysErr: %d",
            localLockRes);
      }
   }

   // flush buffers after we got a new global lock

   if(useGlobalFileLocks &&
      (lockTypeFlags & ENTRYLOCKTYPE_LOCKOPS_ADD) &&
      (globalLockRes == FhgfsOpsErr_SUCCESS) )
   {
      int flushRes = __FhgfsOps_flush(app, file, fhgfs_false, fhgfs_false, fhgfs_true, fhgfs_false);
      if(unlikely(flushRes < 0) )
         return flushRes; // flush error occured
   }

   return FhgfsOpsErr_toSysErr(globalLockRes);
}


ssize_t FhgfsOps_read(struct file* file, char __user *buf, size_t size, loff_t *offsetPointer)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);

   struct inode* inode = file->f_mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   RemotingIOInfo ioInfo;
   ssize_t readRes;

   FhgfsOpsHelper_logOpDebug(app, file->f_dentry, inode, __func__, "(offset: %lld; size: %lld)",
      (long long)*offsetPointer, (long long)size);
   IGNORE_UNUSED_VARIABLE(app);

   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   readRes = FhgfsOpsHelper_readCached(buf, size, *offsetPointer, fhgfsInode, fileInfo, &ioInfo);
   //readRes = FhgfsOpsRemoting_readfile(buf, size, *offsetPointer, &ioInfo);

   if(unlikely(readRes < 0) )
   { // read error (=> transform negative fhgfs error code to system error code)
      return FhgfsOpsErr_toSysErr(-readRes);
   }

   *offsetPointer += readRes;
   FsFileInfo_setLastReadOffset(fileInfo, *offsetPointer);

   if( ( (size_t)readRes < size) && (i_size_read(inode) > *offsetPointer) )
   { // sparse file compatibility mode
      ssize_t readSparseRes = __FhgfsOps_readSparse(
         file, &buf[readRes], size - readRes, *offsetPointer);

      if(unlikely(readSparseRes < 0) )
         return readSparseRes;

      *offsetPointer += readSparseRes;
      readRes += readSparseRes;

      FsFileInfo_setLastReadOffset(fileInfo, *offsetPointer);
   }

   // add to /proc/<pid>/io
   task_io_account_read(readRes);

   return readRes;
}

ssize_t FhgfsOps_readPaged(struct file* file, char __user *buf, size_t size, loff_t* offsetPointer)
{
   ssize_t readRes;
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   struct inode* inode = file->f_mapping->host;

   FhgfsOpsHelper_logOpDebug(app, file->f_dentry, inode, __func__, "(offset: %lld; size: %lld)",
      (long long)*offsetPointer, (long long)size);
   IGNORE_UNUSED_VARIABLE(app);
   IGNORE_UNUSED_VARIABLE(inode);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,16,0)
   readRes = new_sync_read(file, buf, size, offsetPointer); // (will update offsetPointer)
#else
   readRes = do_sync_read(file, buf, size, offsetPointer); // (will update offsetPointer)
#endif // LINUX_VERSION_CODE


   // note: we have no special sparse file read handling here, because that is already included in
      // readpage_sync() and kernel just looks at i_size for file length

   return readRes;
}

/**
 * Read method for hybrid mode.
 * Flush write buffers and call paged read method.
 */
ssize_t FhgfsOps_readPagedHybrid(struct file* file, char __user *buf, size_t size,
   loff_t *offsetPointer)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);

   struct inode* inode = file->f_mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   Mutex* hybridModeMutex;
   Condition* hybridModeCond;
   unsigned short* numHybridReaders;
   unsigned short* numHybridWriters;
   FhgfsOpsErr flushRes;
   ssize_t readRes;

   FhgfsOpsHelper_logOpDebug(app, file->f_dentry, inode, __func__, "(offset: %lld; size: %lld)",
      (long long)*offsetPointer, (long long)size);
   IGNORE_UNUSED_VARIABLE(app);

   FhgfsInode_getHybridModeInfo(fhgfsInode, &hybridModeMutex, &hybridModeCond,
      &numHybridReaders, &numHybridWriters);

   Mutex_lock(hybridModeMutex); // L O C K

   while(*numHybridWriters)
   {
      cond_wait_res_t waitRes = Condition_waitInterruptible(hybridModeCond, hybridModeMutex);
      if(waitRes == COND_WAIT_SIGNAL)
         return -EINTR;
   }

   (*numHybridReaders)++;

   Mutex_unlock(hybridModeMutex); // U N L O C K


   // flush buffer cache
   flushRes = FhgfsOpsHelper_flushCache(app, fhgfsInode, fhgfs_false);
   if(unlikely(flushRes != FhgfsOpsErr_SUCCESS) )
   { // error
      readRes = FhgfsOpsErr_toSysErr(flushRes);
      goto cleanup_and_exit;
   }

   readRes = FhgfsOps_readPaged(file, buf, size, offsetPointer);


cleanup_and_exit:
   Mutex_lock(hybridModeMutex); // L O C K

   (*numHybridReaders)--;

   if(!*numHybridReaders)
      Condition_broadcast(hybridModeCond); // wake up blocked writers

   Mutex_unlock(hybridModeMutex); // U N L O C K

   return readRes;
}

/**
 * Special reading mode that is slower (e.g. not parallel) but compatible with sparse files.
 *
 * Note: Intended to be just a helper for actual read methods (e.g. won't increase the offset
 * pointer).
 *
 * @return negative Linux error code on error, read bytes otherwise
 */
ssize_t __FhgfsOps_readSparse(struct file* file, char __user *buf, size_t size, loff_t offset)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);

   struct inode* inode = file->f_mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   RemotingIOInfo ioInfo;
   ssize_t readRes;
   loff_t i_size;
   FhgfsOpsErr helperReadRes;
   FhgfsIsizeHints iSizeHints;

   FhgfsOpsHelper_logOpDebug(app, file->f_dentry, inode, __func__, "(offset: %lld; size: %lld)",
      (long long)offset, (long long)size);

   readRes = __FhgfsOps_refreshInode(app, inode, NULL, &iSizeHints);
   if(unlikely(readRes) )
      return readRes;

   i_size = i_size_read(inode);
   if(i_size <= offset)
      return 0; // EOF

   // adapt read length to current file length
   size = MIN(size, (unsigned long long)(i_size - offset) );

   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   helperReadRes = FhgfsOpsHelper_readOrClearUser(app, buf, size, offset, fileInfo, &ioInfo);

   if(unlikely(helperReadRes != FhgfsOpsErr_SUCCESS) )
      return FhgfsOpsErr_toSysErr(helperReadRes);

   return size;
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)

ssize_t FhgfsOps_aio_read(struct kiocb *iocb, char __user *buf, size_t count, loff_t pos)
{

#elif LINUX_VERSION_CODE < KERNEL_VERSION(3,16,0)

ssize_t FhgfsOps_aio_read(struct kiocb *iocb, const struct iovec *iov, unsigned long nr_segs,
   loff_t pos)
{
   size_t count = iov_length(iov, nr_segs);

#else // LINUX_VERSION_CODE

ssize_t FhgfsOps_read_iter(struct kiocb *iocb, struct iov_iter *to)
{
   size_t count = iov_iter_count(to);
   loff_t pos = iocb->ki_pos;

#endif // LINUX_VERSION_CODE

   struct file* file = iocb->ki_filp;
   struct address_space* mapping = file->f_mapping;
   struct inode* inode = mapping->host;

   App* app = FhgfsOps_getApp(inode->i_sb);

   ssize_t retVal;

   FhgfsOpsHelper_logOpDebug(app, file->f_dentry, inode, __func__, "(offset: %lld; size: %lld)",
      (long long)pos, (long long)count);

   IGNORE_UNUSED_VARIABLE(pos);
   IGNORE_UNUSED_VARIABLE(count);

   retVal = __FhgfsOps_revalidateMapping(app, inode);
   if(unlikely(retVal) )
   { // error
      return retVal;
   }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
   retVal = generic_file_aio_read(iocb, buf, count, pos);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3,16,0)
   retVal = generic_file_aio_read(iocb, iov, nr_segs, pos);
#else // LINUX_VERSION_CODE
   retVal = generic_file_read_iter(iocb, to);
#endif // LINUX_VERSION_CODE

   return retVal;
}


ssize_t FhgfsOps_write(struct file* file, const char __user *buf, size_t size,
   loff_t* offsetPointer)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Config* cfg = App_getConfig(app);

   int writeCheckRes;
   struct inode* inode = file->f_mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   RemotingIOInfo ioInfo;
   loff_t writeOffset;
   ssize_t writeRes;
   loff_t newMinFileSize; // to update i_size after write

   fhgfs_bool isLocallyLockedAppend =
      FsFileInfo_getAppending(fileInfo) && !Config_getTuneUseGlobalAppendLocks(cfg);
   fhgfs_bool isGloballyLockedAppend =
      FsFileInfo_getAppending(fileInfo) && Config_getTuneUseGlobalAppendLocks(cfg);

   FhgfsOpsHelper_logOpDebug(app, file->f_dentry, inode, __func__, "(offset: %lld; size: %lld)",
      (long long)*offsetPointer, (long long)size);

   writeCheckRes = generic_write_checks(file, offsetPointer, &size, S_ISBLK(inode->i_mode) );
   if(unlikely(writeCheckRes) )
      return writeCheckRes;

   if(isLocallyLockedAppend)
   { // appending without global locks => move file offset to end-of-file before writing

      /* note on flush and lock: the flush here must be inside the local lock, but cannot happen at
         the place where we take the global lock (because that might be called from a flush path
         itself), that's why global and local locks are taken at different places. */

      int flushRes;
      FhgfsOpsErr statRes;
      fhgfs_stat fhgfsStat;

      Fhgfsinode_appendLock(fhgfsInode); // L O C K (append)

      flushRes = __FhgfsOps_flush(app, file, fhgfs_false, fhgfs_false, fhgfs_true, fhgfs_false);
      if(unlikely(flushRes < 0) )
      { // flush error
         writeRes = flushRes;
         goto unlockappend_and_exit;
      }

      FhgfsInode_entryInfoReadLock(fhgfsInode); // LOCK EntryInfo

      /* note on stat here: we could pass -1 to _writeCached and remove the stat here, but the
         disadvantage would be that we don't have the correct file offset for i_size then, so we
         leave the stat here for now. */

      statRes = FhgfsOpsRemoting_statDirect(app, FhgfsInode_getEntryInfo(fhgfsInode), &fhgfsStat);

      FhgfsInode_entryInfoReadUnlock(fhgfsInode); // UNLOCK EntryInfo

      if(unlikely(statRes != FhgfsOpsErr_SUCCESS) )
      { // remote stat error
         writeRes = FhgfsOpsErr_toSysErr(statRes);
         goto unlockappend_and_exit;
      }

      *offsetPointer = fhgfsStat.size;
   }

   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   writeOffset = isGloballyLockedAppend ? -1 : *offsetPointer;

   writeRes = FhgfsOpsHelper_writeCached(buf, size, writeOffset, fhgfsInode, fileInfo, &ioInfo);
   //writeRes = FhgfsOpsRemoting_writefile(buf, size, *offsetPointer, &ioInfo);

   if(unlikely(writeRes < 0) )
   { // write error (=> transform negative fhgfs error code to system error code)
      writeRes = FhgfsOpsErr_toSysErr(-writeRes);
      goto unlockappend_and_exit;
   }

   if(!isGloballyLockedAppend)
   { // for (buffered) global append locks, new offset/filesize would be unknown
      newMinFileSize = *offsetPointer + writeRes; // update with old offset to avoid offset==0 check
      *offsetPointer += writeRes;

      FsFileInfo_setLastWriteOffset(fileInfo, *offsetPointer);

      // check current file size and update if necessary (also important for sparse read heuristic)
      spin_lock(&inode->i_lock);
      if(inode->i_size < newMinFileSize)
         i_size_write(inode, newMinFileSize);
      spin_unlock(&inode->i_lock);
   }

   // add to /proc/<pid>/io
   task_io_account_write(writeRes);

unlockappend_and_exit:
   if(isLocallyLockedAppend)
      Fhgfsinode_appendUnlock(fhgfsInode); // U N L O C K (append)

   return writeRes;
}

/**
 * Write method for hybrid cache mode.
 * Flush read pages and call buffered write method.
 */
ssize_t FhgfsOps_writeHybrid(struct file* file, const char __user *buf, size_t size,
   loff_t* offsetPointer)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);

   struct inode* inode = file->f_mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   Mutex* hybridModeMutex;
   Condition* hybridModeCond;
   unsigned short* numHybridReaders;
   unsigned short* numHybridWriters;
   ssize_t writeRes;

   FhgfsOpsHelper_logOpDebug(app, file->f_dentry, inode, __func__, "(offset: %lld; size: %lld)",
      (long long)*offsetPointer, (long long)size);
   IGNORE_UNUSED_VARIABLE(app);

   FhgfsInode_getHybridModeInfo(fhgfsInode, &hybridModeMutex, &hybridModeCond,
      &numHybridReaders, &numHybridWriters);


   Mutex_lock(hybridModeMutex); // L O C K

   while(*numHybridReaders)
   {
      cond_wait_res_t waitRes = Condition_waitInterruptible(hybridModeCond, hybridModeMutex);
      if(waitRes == COND_WAIT_SIGNAL)
         return -EINTR;
   }

   (*numHybridWriters)++;

   Mutex_unlock(hybridModeMutex); // U N L O C K


   invalidate_remote_inode(inode); // invalidate read pages

   writeRes = FhgfsOps_write(file, buf, size, offsetPointer);


   Mutex_lock(hybridModeMutex); // L O C K

   (*numHybridWriters)--;

   if(!*numHybridWriters)
      Condition_broadcast(hybridModeCond); // wake up blocked readers

   Mutex_unlock(hybridModeMutex); // U N L O C K

   return writeRes;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)

ssize_t FhgfsOps_aio_write(struct kiocb *iocb, const char __user *buf, size_t count, loff_t pos)
{

#elif LINUX_VERSION_CODE < KERNEL_VERSION(3,16,0)

ssize_t FhgfsOps_aio_write(struct kiocb *iocb, const struct iovec *iov, unsigned long nr_segs,
   loff_t pos)
{
   size_t count = iov_length(iov, nr_segs);

#else

ssize_t FhgfsOps_write_iter(struct kiocb *iocb, struct iov_iter *from)
{
   size_t count = iov_iter_count(from);
   loff_t pos = iocb->ki_pos;

#endif // LINUX_VERSION_CODE

   struct file* file = iocb->ki_filp;
   struct dentry* dentry = file->f_dentry;
   struct inode* inode = file_inode(file);
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = __func__;

   FhgfsIsizeHints iSizeHints;

   ssize_t retVal;
   int writeCheckRes;

   FhgfsOpsHelper_logOpDebug(app, dentry, inode, __func__, "(offset: %lld; size: %lld)",
      (long long)pos, (long long)count);

   if (iocb->ki_pos != pos)
   { /* Similiar to WARN_ON(iocb->ki_pos != pos), as fuse does */
      Logger_logErrFormatted(log, logContext, "Bug: iocb->ki_pos != pos (%ld vs %ld)",
         iocb->ki_pos, pos);

      dump_stack();
   }

   if(iocb->ki_filp->f_flags & O_APPEND)
   { // O_APPEND => flush (for correct size) and refresh file size

      retVal = __FhgfsOps_refreshInode(app, inode, NULL, &iSizeHints);
      if(retVal)
         return retVal;
   }

   writeCheckRes = generic_write_checks(file, &pos, &count, S_ISBLK(inode->i_mode) );
   if(unlikely(writeCheckRes) )
      return writeCheckRes;

   if(!count)
      return 0;

   if (file->f_flags & O_APPEND && pos != iocb->ki_pos)
   {
      /* pos was updated by generic_write_checks (append writes), so we also need to update
       * iocb->ki_pos, otherwise generic_file_aio_write() will call BUG_ON */
      iocb->ki_pos = pos;
   }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
   retVal = generic_file_aio_write(iocb, buf, count, pos);
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3,16,0)
   retVal = generic_file_aio_write(iocb, iov, nr_segs, pos);
#else
   iov_iter_truncate(from, count);

   retVal = generic_file_write_iter(iocb, from);
#endif // LINUX_VERSION_CODE

   if( (retVal >= 0)
      && ( (IS_SYNC(inode) || (iocb->ki_filp->f_flags & O_SYNC) )
         || unlikely(FhgfsInode_getHasWritePageError(fhgfsInode)) ) )
   { // sync I/O => flush and wait
      struct address_space* mapping = inode->i_mapping;

      if(mapping->nrpages)
      {
         int writeRes = filemap_fdatawrite(mapping);
         if(writeRes >= 0)
         {
            int waitRes = filemap_fdatawait(mapping);
            if(waitRes < 0)
               retVal = waitRes;
         }
         else
            retVal = writeRes;
      }

      if (unlikely(FhgfsInode_getHasWritePageError(fhgfsInode) ) && retVal >= 0)
         FhgfsInode_clearWritePageError(fhgfsInode);
   } // end of if(sync)

   return retVal;
}


#ifdef KERNEL_HAS_FSYNC_RANGE /* added in vanilla 3.1 */
int FhgfsOps_fsync(struct file* file, loff_t start, loff_t end, int datasync)
{
   struct dentry* dentry = file->f_dentry;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35)
int FhgfsOps_fsync(struct file* file, int datasync)
{
      struct dentry* dentry = file->f_dentry;
#else /* LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,34) */
int FhgfsOps_fsync(struct file* file, struct dentry* dentry, int datasync)
{
#endif // LINUX_VERSION_CODE
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Config* cfg = App_getConfig(app);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_fsync";

   int retVal = 0;
   FsObjectInfo* fsObjectInfo;

   struct inode* inode = file_inode(file);

   if(unlikely(Logger_getLogLevel(log) >= 5) )
      FhgfsOpsHelper_logOp(5, app, dentry, inode, logContext);

   fsObjectInfo = __FhgfsOps_getObjectInfo(file);

   // syncing something other than a file?
   if(FsObjectInfo_getObjectType(fsObjectInfo) != FsObjectType_FILE)
      goto clean_up;

   // TODO: Implement start and end here as allowed by >linux-3.1

   retVal = __FhgfsOps_flush(app, file, fhgfs_false, Config_getTuneRemoteFSync(cfg), fhgfs_true,
      fhgfs_false);

   if(retVal)
      goto clean_up;


clean_up:

   return retVal;
}

/**
 * Flush data from local cache to servers.
 * Note: This method is in fact a local sync (as wait until data have arrived remotely)
 *
 * @param discardCacheOnError whether or not to discard the internal buffer cache in case of an
 * error (typically you will only want to set fhgfs_true here during file close).
 * @param forceRemoteFlush whether or not remote fsync will be executed depends on this value
 * and the corresponding config value (fsync will use fhgfs_true here and flush won't)
 * @param checkSession whether or not a server crash detection must be done on the server
 * @param isClose whether or not the method is called by a close
 * @return negative linux error code on error
 */
int __FhgfsOps_flush(App* app, struct file *file, fhgfs_bool discardCacheOnError,
   fhgfs_bool forceRemoteFlush, fhgfs_bool checkSession, fhgfs_bool isClose)
{
   Logger* log = App_getLogger(app);
   Config* cfg = App_getConfig(app);
   const char* logContext = __func__;

   struct inode* inode = file->f_mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   RemotingIOInfo ioInfo;
   int filemapWaitRes;
   int inodeWriteRes;
   FhgfsOpsErr flushRes;
   int retVal = 0;
   fhgfs_bool hasWriteHandle = FhgfsInode_hasWriteHandle(fhgfsInode);
   
   fhgfs_bool doSyncOnClose = Config_getSysSyncOnClose(App_getConfig(app)) && isClose;

   if(unlikely(Logger_getLogLevel(log) >= 5) )
      FhgfsOpsHelper_logOp(Log_SPAM, app, file->f_dentry, inode, logContext);

   if (hasWriteHandle || FhgfsInode_getHasDirtyPages(fhgfsInode) )
   {
      // flush page cache
      inodeWriteRes = write_inode_now(inode, 1);
      filemapWaitRes = filemap_fdatawait(file->f_mapping);
      if(unlikely(inodeWriteRes < 0 || filemapWaitRes < 0) )
      {
         retVal = (inodeWriteRes < 0) ? inodeWriteRes : filemapWaitRes;
         goto clean_up;
      }
   }

   // flush buffer cache

   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   flushRes = FhgfsOpsHelper_flushCache(app, fhgfsInode, discardCacheOnError);
   if(unlikely(flushRes != FhgfsOpsErr_SUCCESS) )
   { // error
      retVal = FhgfsOpsErr_toSysErr(flushRes);
      goto clean_up;
   }

   // remote fsync

   if(forceRemoteFlush || (checkSession && Config_getSysSessionCheckOnClose(cfg)) || doSyncOnClose)
   {
      FhgfsOpsErr fsyncRes = FhgfsOpsRemoting_fsyncfile(&ioInfo, forceRemoteFlush, checkSession,
         doSyncOnClose);
      if(unlikely(fsyncRes != FhgfsOpsErr_SUCCESS) )
      {
         retVal = FhgfsOpsErr_toSysErr(fsyncRes);
         goto clean_up;
      }
   }

clean_up:

   return retVal;
}

int FhgfsOps_mmap(struct file* file, struct vm_area_struct* vma)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_mmap";

   FhgfsIsizeHints iSizeHints;

   int retVal;
   struct inode* inode = file_inode(file);

   if(unlikely(Logger_getLogLevel(log) >= 5) )
      FhgfsOpsHelper_logOp(5, app, file->f_dentry, inode, logContext);

   retVal = generic_file_mmap(file, vma);
   if(!retVal)
   {
      retVal = __FhgfsOps_refreshInode(app, inode, NULL, &iSizeHints);
   }

   LOG_DEBUG_FORMATTED(log, 5, logContext, "result: %d", retVal);

   return retVal;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,28)

/**
 * Note: We never return an error code here, currently.
 *
 * @param from page inner offset start
 * @param to page inner offset end
 * @return 0 on success
 */
int FhgfsOps_prepare_write(struct file* file, struct page* page, unsigned from, unsigned to)
{
   const char* logContext = __func__;

   struct inode* inode = file->f_mapping->host;

   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);

   int retVal = 0;

   loff_t fileSize;
   loff_t offset;

   FhgfsOpsHelper_logOpDebug(app, file->f_dentry, inode, logContext, "from: %u; to: %u", from, to);
   IGNORE_UNUSED_VARIABLE(logContext);
   IGNORE_UNUSED_VARIABLE(inode);

   if(PageUptodate(page) )
      goto clean_up;

   if( (to == PAGE_CACHE_SIZE) && (from == 0) )
   { // we're writing a full page
      // full page => will be up to date => no need to read-update the page from the server
      SetPageUptodate(page);
      goto clean_up;
   }

   // we're not writing a full page => try to read-update the page

   offset = page_offset(page);
   fileSize = i_size_read(page->mapping->host);

   LOG_DEBUG_FORMATTED(log, 5, logContext, "fileSize: %lld; offset: %lld",
      (long long)fileSize, (long long)offset);
   IGNORE_UNUSED_VARIABLE(log);

   if( (offset >= fileSize) ||
       ( (from == 0) && (offset + to) >= fileSize) )
   {
      // we don't need to read data beyond the end of the file
      //    => zero it, and set the page uptodate
      simple_prepare_write(file, page, from, to);
      SetPageUptodate(page);
   }
   else
   if( (file->f_flags & O_ACCMODE) != O_WRONLY)
   {
      // file open for reading also => read-update the page
      retVal = FhgfsOpsPages_readpageSync(file, page);
   }
   else
   { // Write-only file handle => we cannot use it to read-update this page.
      // We could try to get a different open file handle, but the user data will be written out by
      // commit_write, so it's fine.
   }


   // clean-up
clean_up:

   IGNORE_UNUSED_VARIABLE(retVal);
   return 0;
}

/**
 * @param from page inner offset start
 * @param to page inner offset end
 *
 * @return 0 in success
 */
int FhgfsOps_commit_write(struct file* file, struct page* page, unsigned from, unsigned to)
{
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = "FhgfsOps_commit_write";

   int retVal = 0;
   struct inode* inode = file->f_mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   loff_t pageStart = page_offset(page);
   loff_t endPos = pageStart + to;

   FhgfsOpsHelper_logOpDebug(app, file->f_dentry, inode, __func__,
      "pageStart: %lld; from: %u; to: %u", (long long)pageStart, from, to);

   // check fileSize and update it to page endPos if required
   spin_lock(&inode->i_lock);
   if(endPos > inode->i_size)
   {
      FhgfsInode_setLastWriteBackOrIsizeWriteTime(fhgfsInode);
      FhgfsInode_setPageWriteFlag(fhgfsInode);
      i_size_write(inode, endPos);
   }
   spin_unlock(&inode->i_lock);

   // write the page directly or mark it for async writing
   if (unlikely(!PageUptodate(page) ) )
      retVal = -EIO; // something failed on reading the page, tell the user about it
   else
   {  // mark for async writing via writepage
      if (!PageDirty(page) )
      { // Only add if the page is not dirty yet (don't add the same page twice...)
         FhgfsInode_incNumDirtyPages(fhgfsInode);
      }
      set_page_dirty(page);
   }


   LOG_DEBUG_FORMATTED(log, 5, logContext, "complete. retVal: %d", retVal);
   IGNORE_UNUSED_VARIABLE(log);
   IGNORE_UNUSED_VARIABLE(logContext);

   return retVal;
}

#else // LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28)

/**
 * @param fsdata can be used to hand any data over to write_end() (but note that the old
 * prepare_write() doesn't have this)
 * @return 0 on success
 */
int FhgfsOps_write_begin(struct file* file, struct address_space* mapping,
   loff_t pos, unsigned len, unsigned flags, struct page** pagep, void** fsdata)
{
   pgoff_t index = pos >> PAGE_CACHE_SHIFT;
   loff_t offset = pos & (PAGE_CACHE_SIZE - 1);
   loff_t page_start = pos & PAGE_MASK;

   loff_t i_size;
   struct page* page;

   int retVal = 0;
   App* app = FhgfsOps_getApp(file->f_dentry->d_sb);

   // FhgfsOpsHelper_logOpDebug(app, file->f_dentry, __func__, "(offset: %lld; page_start: %lld; len: %u)",
   //   (long long)offset, (long long)page_start, len);
   IGNORE_UNUSED_VARIABLE(app);

   page = grab_cache_page_write_begin(mapping, index, flags);
   if(!page)
   {
      retVal = -ENOMEM;
      goto clean_up;
   }

   if(PageUptodate(page) )
      goto clean_up;

   if(len == PAGE_CACHE_SIZE)
   {
      // two possibilities:
      // a) full page write => no need to read-update the page from the server
      // b) short write (with offset) => will lead to sync write
      goto clean_up;
   }

   i_size = i_size_read(mapping->host);
   if( (page_start >= i_size) ||
       (!offset && ( (pos + len) >= i_size) ) )
   {
      // we don't need to read data beyond the end of the file
      //    => zero it, and set the page up-to-date

      zero_user_segments(page, 0, offset, offset + len, PAGE_CACHE_SIZE);

      // note: PageChecked means rest of the page (to which is not being written) is up-to-date.
      //       so when our data is written, the whole page is up-to-date.
      SetPageChecked(page);

      goto clean_up;
   }

   // it is read-modify-write, so update the page with the server content
   retVal = FhgfsOpsPages_readpageSync(file, page);

clean_up:
   // clean-up

   *pagep = page;

   return retVal;
}

/**
 * @param copied the amount that was able to be copied ("copied==len" is always true if
 * write_begin() was called with the AOP_FLAG_UNINTERRUPTIBLE flag)
 * @param fsdata whatever write_begin() set here
 * @return < 0 on failure, number of bytes copied into pagecache (<= 'copied') otherwise
 */
int FhgfsOps_write_end(struct file* file, struct address_space* mapping,
   loff_t pos, unsigned len, unsigned copied, struct page* page, void* fsdata)
{
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);

   struct inode* inode = mapping->host;
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);

   int retVal;

   struct dentry *dentry= file->f_dentry;
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);
   const char* logContext = __func__;

   // FhgfsOpsHelper_logOpDebug(app, dentry, logContext, "pos: %lld; len: %u; copied: %u",
   //   (long long)pos, len, copied);
   IGNORE_UNUSED_VARIABLE(logContext);

   if(PageChecked(page) )
   {
      // note: see write_begin() for meaning of PageChecked()

      if(copied == len)
         SetPageUptodate(page);

      ClearPageChecked(page);
   }
   else
   if(!PageUptodate(page) && (copied == PAGE_CACHE_SIZE) )
      SetPageUptodate(page);


   if(!PageUptodate(page) )
   {
      unsigned offset = pos & (PAGE_CACHE_SIZE - 1);
      char* buf = kmap(page);
      RemotingIOInfo ioInfo;
      ssize_t writeRes;

      FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

      FhgfsInode_incWriteBackCounter(fhgfsInode);

      writeRes = FhgfsOpsRemoting_writefile(&buf[offset], copied, pos, &ioInfo);

      spin_lock(&inode->i_lock);
      FhgfsInode_setLastWriteBackOrIsizeWriteTime(fhgfsInode);
      FhgfsInode_decWriteBackCounter(fhgfsInode);
      spin_unlock(&inode->i_lock);

      if(likely(writeRes > 0) )
      {
         retVal = writeRes;
         pos += writeRes;
      }
      else
         retVal = FhgfsOpsErr_toSysErr(-writeRes);

      kunmap(page);
   }
   else
   {
      retVal = copied;
      pos += copied;

      if (!PageDirty(page) )
      {  // Only add if the page is not dirty yet (don't add the same page twice...)
         FhgfsInode_incNumDirtyPages(fhgfsInode);
      }
      set_page_dirty(page); // could be in the if-condition above, but for safety we set it here

   }

   if(likely(retVal > 0) )
   {
      spin_lock(&inode->i_lock);
      if(pos > inode->i_size)
      {
         FhgfsInode_setPageWriteFlag(fhgfsInode);
         FhgfsInode_setLastWriteBackOrIsizeWriteTime(fhgfsInode);
         FhgfsInode_setNoIsizeDecrease(fhgfsInode);
         i_size_write(inode, pos);
      }
      spin_unlock(&inode->i_lock);
   }

   unlock_page(page);
   page_cache_release(page);

   // clean-up

   // LOG_DEBUG_FORMATTED(log, 5, logContext, "complete. retVal: %d", retVal);
   IGNORE_UNUSED_VARIABLE(log);

   return retVal;
}

#endif // LINUX_VERSION_CODE


/**
 * Note: This method must be defined because otherwise the kernel rejects open() with O_DIRECT in
 * fs/open.c. However, it is only called indirectoy through the generic file read/write routines
 * (and swapping code), so it should actually never be called for buffered IO.
 *
 * @param rw whether this is a read or a write {READ, WRITE}
 * @param iocb I/O control block with open file handle
 * @param iov I/O buffer vectors (array)
 * @param pos file offset
 * @param nr_segs length of iov array
 */
#ifdef KERNEL_HAS_DIRECT_IO_ITER

ssize_t FhgfsOps_directIO(int rw, struct kiocb *iocb, struct iov_iter *iter, loff_t pos)
{

   #ifdef KERNEL_HAS_SPECIAL_UEK_IOV_ITER
      // oracle uek seems to be the only kernel with a special version of struct iov_iter
      const struct iovec* iov = iov_iter_iovec(iter);
   #else // KERNEL_HAS_SPECIAL_UEK_IOV_ITER
      const struct iovec* iov = iter->iov;
   #endif // KERNEL_HAS_SPECIAL_UEK_IOV_ITER

   unsigned long nr_segs = iter->nr_segs;

#else // KERNEL_HAS_DIRECT_IO_ITER

ssize_t FhgfsOps_directIO(int rw, struct kiocb *iocb, const struct iovec *iov, loff_t pos,
   unsigned long nr_segs)
{

#endif // KERNEL_HAS_DIRECT_IO_ITER

   struct file* file = iocb->ki_filp;
   FsFileInfo* fileInfo = __FhgfsOps_getFileInfo(file);
   struct dentry* dentry = file->f_dentry;
   struct inode* inode = file_inode(file);
   FhgfsInode* fhgfsInode = FHGFS_INODE(inode);
   RemotingIOInfo ioInfo;

   ssize_t remotingRes;
   ssize_t retVal = 0;
   
   unsigned long i;

   const char* logContext = __func__;
   App* app = FhgfsOps_getApp(dentry->d_sb);
   Logger* log = App_getLogger(app);

   FhgfsOpsHelper_logOpDebug(app, dentry, inode, logContext, "(%s, pos: %lld, nr_seqs: %lld)",
      (rw == WRITE) ? "WRITE" : "READ", (long long)pos, (long long)nr_segs);
   IGNORE_UNUSED_VARIABLE(logContext); // non-debug builds

   FsFileInfo_getIOInfo(fileInfo, fhgfsInode, &ioInfo);

   for(i=0; i < nr_segs; i++)
   {
      const struct iovec* currentIOV = &iov[i];

      if(rw == WRITE)
      { // write
         remotingRes = FhgfsOpsRemoting_writefile(
            currentIOV->iov_base, currentIOV->iov_len, pos, &ioInfo);
      }
      else
      if(rw == READ)
      { // read
         remotingRes = FhgfsOpsRemoting_readfile(
            currentIOV->iov_base, currentIOV->iov_len, pos, &ioInfo);

         /* note: casting remotingRes below is required to be "gcc -Wsign-compare" clean).
            (casting potentially negative remotingRes to unsigned doesn't matter, because we also
            have a remotingRes>=0 check in the if-statement.) */

         if( ( (size_t)remotingRes < currentIOV->iov_len) &&
             (remotingRes >= 0) &&
             ( i_size_read(inode) > (pos + (loff_t) currentIOV->iov_len) ) )
         { // sparse file compatibility mode
            ssize_t readSparseRes = __FhgfsOps_readSparse(file, currentIOV->iov_base + remotingRes,
               currentIOV->iov_len - remotingRes, pos + currentIOV->iov_len);

            if(unlikely(readSparseRes < 0) )
            {
               retVal = readSparseRes;
               break;
            }

            remotingRes += readSparseRes;
         }
      }
      else
      { // unhandled case

         #ifdef WARN_ONCE
         WARN_ONCE(1, "unexpected: rw value !=READ and !=WRITE. (int value: %d)\n", rw);
         #endif

         return -EINVAL;
      }

      if(unlikely(remotingRes < 0) )
      { // error occurred
         LOG_DEBUG_FORMATTED(log, 1, logContext, "error: %s",
            FhgfsOpsErr_toErrString(-remotingRes) );
         IGNORE_UNUSED_VARIABLE(log);

         retVal = FhgfsOpsErr_toSysErr(-remotingRes);
         break;
      }

      retVal += remotingRes;
      pos += remotingRes;

      if(remotingRes != (ssize_t)currentIOV->iov_len)
         break; // read end of file or no write space left
   }


   if(likely(retVal > 0) )
   { // add to /proc/<pid>/io
      if(rw == WRITE)
         task_io_account_write(retVal);
      else
         task_io_account_read(retVal);
   }

   return retVal;
}



